from PyQt5.QtCore import Qt, QMetaObject, pyqtSignal, pyqtSlot
from PyQt5.QtWidgets import (QWidget, QVBoxLayout, QHBoxLayout, QToolButton, QLabel, QSizePolicy, QDialog, QMessageBox)
from ._utils import QT_VERSION, PLATFORM, resource_path


_FL_STYLESHEET = resource_path('resources/frameless.qss')
""" str: Frameless window stylesheet. """


class WindowDragger(QWidget):
    """ Window dragger.

        Args:
            window (QWidget): Associated window.
            parent (QWidget, optional): Parent widget.
    """

    doubleClicked = pyqtSignal()

    def __init__(self, window, parent=None):
        QWidget.__init__(self, parent)

        self._window = window
        self._mousePressed = False

    def mousePressEvent(self, event):
        self._mousePressed = True
        self._mousePos = event.globalPos()
        self._windowPos = self._window.pos()

    def mouseMoveEvent(self, event):
        if self._mousePressed:
            self._window.move(self._windowPos +
                              (event.globalPos() - self._mousePos))

    def mouseReleaseEvent(self, event):
        self._mousePressed = False

    def mouseDoubleClickEvent(self, event):
        self.doubleClicked.emit()


class TitleBar(WindowDragger):
    """ Window titlebar.
        Args:
            window (QWidget): The window to attach to.
    """
    def __init__(self, window):
        super().__init__(window, window.windowFrame)
        self.window = window
        self.setupUi()

    def setupUi(self):
        self.setObjectName('titleBar')
        self.setSizePolicy(QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed))

        self.window.hboxTitle = QHBoxLayout(self)
        self.window.hboxTitle.setContentsMargins(5, 0, 0, 0)
        self.window.hboxTitle.setSpacing(0)

        self.window.lblTitle = QLabel('Title')
        self.window.lblTitle.setObjectName('lblTitle')
        self.window.lblTitle.setAlignment(Qt.AlignCenter)

        spButtons = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)

        self.window.btnMinimize = QToolButton(self)
        self.window.btnMinimize.setObjectName('btnMinimize')
        self.window.btnMinimize.setSizePolicy(spButtons)

        self.window.btnRestore = QToolButton(self)
        self.window.btnRestore.setObjectName('btnRestore')
        self.window.btnRestore.setSizePolicy(spButtons)

        self.window.btnMaximize = QToolButton(self)
        self.window.btnMaximize.setObjectName('btnMaximize')
        self.window.btnMaximize.setSizePolicy(spButtons)

        self.window.btnClose = QToolButton(self)
        self.window.btnClose.setObjectName('btnClose')
        self.window.btnClose.setSizePolicy(spButtons)

        self.window.vboxFrame.addWidget(self)

        self.window.windowContent = QWidget(self.window.windowFrame)
        self.window.vboxFrame.addWidget(self.window.windowContent)

        self.window.vboxWindow.addWidget(self.window.windowFrame)

        if PLATFORM == "Darwin":
            self.window.hboxTitle.addWidget(self.window.btnClose)
            self.window.hboxTitle.addWidget(self.window.btnMinimize)
            self.window.hboxTitle.addWidget(self.window.btnRestore)
            self.window.hboxTitle.addWidget(self.window.btnMaximize)
            self.window.hboxTitle.addWidget(self.window.lblTitle)
        else:
            self.window.hboxTitle.addWidget(self.window.lblTitle)
            self.window.hboxTitle.addWidget(self.window.btnMinimize)
            self.window.hboxTitle.addWidget(self.window.btnRestore)
            self.window.hboxTitle.addWidget(self.window.btnMaximize)
            self.window.hboxTitle.addWidget(self.window.btnClose)


class ModernWindow(QWidget):
    """ Modern window.

        Args:
            w (QWidget): Main widget.
            parent (QWidget, optional): Parent widget.
    """

    def __init__(self, w, hide_disabled_btns=True, parent=None):
        QWidget.__init__(self, parent)

        self._w = w
        self._hide_disabled_btns = hide_disabled_btns
        self.setupUi()

        contentLayout = QHBoxLayout()
        contentLayout.setContentsMargins(0, 0, 0, 0)
        contentLayout.addWidget(w)

        self.windowContent.setLayout(contentLayout)

        self.setWindowTitle(w.windowTitle())
        self.setGeometry(w.geometry())

        # Adding attribute to clean up the parent window when the child is closed
        # self._w.setAttribute(Qt.WA_DeleteOnClose, True)
        # self._w.destroyed.connect(self.__child_was_closed)

    def setupUi(self):
        # create title bar, content
        self.vboxWindow = QVBoxLayout(self)
        self.vboxWindow.setContentsMargins(0, 0, 0, 0)

        self.windowFrame = QWidget(self)
        self.windowFrame.setObjectName('windowFrame')

        self.vboxFrame = QVBoxLayout(self.windowFrame)
        self.vboxFrame.setContentsMargins(0, 0, 0, 0)

        self.titleBar = TitleBar(self)

        # set window flags
        self.setWindowFlags(Qt.Window | Qt.FramelessWindowHint | Qt.WindowSystemMenuHint | Qt.WindowCloseButtonHint |
                            Qt.WindowMinimizeButtonHint | Qt.WindowMaximizeButtonHint)

        if QT_VERSION >= (5,):
            self.setAttribute(Qt.WA_TranslucentBackground)

        # set stylesheet
        with open(_FL_STYLESHEET) as stylesheet:
            self.setStyleSheet(stylesheet.read())

        # automatically connect slots
        QMetaObject.connectSlotsByName(self)

    def __child_was_closed(self):
        self._w = None  # The child was deleted, remove the reference to it and close the parent window
        self.close()

    def show(self):
        self._w.show()
        super().show()

    def closeEvent(self, event):
        if not self._w:
            event.accept()
        else:
            self._w.close()
            # event.setAccepted(self._w.isHidden())
            event.accept()

    def setWindowTitle(self, title):
        """ Set window title.

            Args:
                title (str): Title.
        """

        super(ModernWindow, self).setWindowTitle(title)
        self.lblTitle.setText(title)

    def _setWindowButtonState(self, hint, state):
        btns = {
            Qt.WindowCloseButtonHint: self.btnClose,
            Qt.WindowMinimizeButtonHint: self.btnMinimize,
            Qt.WindowMaximizeButtonHint: self.btnMaximize
        }
        button = btns.get(hint)

        maximized = bool(self.windowState() & Qt.WindowMaximized)

        if button == self.btnMaximize:  # special rules for max/restore
            self.btnRestore.setEnabled(state)
            self.btnMaximize.setEnabled(state)

            if maximized:
                self.btnRestore.setVisible(state)
                self.btnMaximize.setVisible(False)
            else:
                self.btnMaximize.setVisible(state)
                self.btnRestore.setVisible(False)
        else:
            button.setEnabled(state)

        allButtons = [self.btnClose, self.btnMinimize, self.btnMaximize, self.btnRestore]
        if True in [b.isEnabled() for b in allButtons]:
            for b in allButtons:
                if self._hide_disabled_btns:
                    b.setVisible(b.isEnabled())
                else:
                    b.setVisible(True)
            if maximized:
                self.btnMaximize.setVisible(False)
            else:
                self.btnRestore.setVisible(False)
            self.lblTitle.setContentsMargins(0, 0, 0, 0)
        else:
            for b in allButtons:
                b.setVisible(False)
            self.lblTitle.setContentsMargins(0, 2, 0, 0)

    def setWindowFlag(self, Qt_WindowType, on=True):
        buttonHints = [Qt.WindowCloseButtonHint, Qt.WindowMinimizeButtonHint, Qt.WindowMaximizeButtonHint]

        if Qt_WindowType in buttonHints:
            self._setWindowButtonState(Qt_WindowType, on)
        else:
            QWidget.setWindowFlag(self, Qt_WindowType, on)

    def setWindowFlags(self, Qt_WindowFlags):
        buttonHints = [Qt.WindowCloseButtonHint, Qt.WindowMinimizeButtonHint, Qt.WindowMaximizeButtonHint]
        for hint in buttonHints:
            self._setWindowButtonState(hint, bool(Qt_WindowFlags & hint))

        QWidget.setWindowFlags(self, Qt_WindowFlags)

    @pyqtSlot()
    def on_btnMinimize_clicked(self):
        self.setWindowState(Qt.WindowMinimized)

    @pyqtSlot()
    def on_btnRestore_clicked(self):
        if self.btnMaximize.isEnabled() or self.btnRestore.isEnabled():
            self.btnRestore.setVisible(False)
            self.btnRestore.setEnabled(False)
            self.btnMaximize.setVisible(True)
            self.btnMaximize.setEnabled(True)

        self.setWindowState(Qt.WindowNoState)

    @pyqtSlot()
    def on_btnMaximize_clicked(self):
        if self.btnMaximize.isEnabled() or self.btnRestore.isEnabled():
            self.btnRestore.setVisible(True)
            self.btnRestore.setEnabled(True)
            self.btnMaximize.setVisible(False)
            self.btnMaximize.setEnabled(False)

        self.setWindowState(Qt.WindowMaximized)

    @pyqtSlot()
    def on_btnClose_clicked(self):
        self.close()

    @pyqtSlot()
    def on_titleBar_doubleClicked(self):
        if not bool(self.windowState() & Qt.WindowMaximized):
            self.on_btnMaximize_clicked()
        else:
            self.on_btnRestore_clicked()


class ModernDialog(QDialog):
    """ Modern window.

        Args:
            w (QDialog): Main widget.
            parent (QWidget, optional): Parent widget.
    """

    def __init__(self, parent=None):
        QDialog.__init__(self, parent)
        self.setupUi()

    def setupUi(self):
        # create title bar, content
        self.vboxWindow = QVBoxLayout(self)
        self.vboxWindow.setContentsMargins(0, 0, 0, 0)

        self.windowFrame = QWidget(self)
        self.windowFrame.setObjectName('windowFrame')

        self.vboxFrame = QVBoxLayout(self.windowFrame)
        self.vboxFrame.setContentsMargins(0, 0, 0, 0)

        self.titleBar = TitleBar(self)

        # set window flags
        self.setWindowFlags(Qt.Window | Qt.FramelessWindowHint | Qt.WindowSystemMenuHint | Qt.WindowCloseButtonHint |
                            Qt.WindowMinimizeButtonHint | Qt.WindowMaximizeButtonHint)

        if QT_VERSION >= (5,):
            self.setAttribute(Qt.WA_TranslucentBackground)

        # set stylesheet
        with open(_FL_STYLESHEET) as stylesheet:
            self.setStyleSheet(stylesheet.read())

        # automatically connect slots
        QMetaObject.connectSlotsByName(self)

    def closeEvent(self, event):
        event.accept()

    def setWindowTitle(self, title):
        """ Set window title.

            Args:
                title (str): Title.
        """

        super(ModernDialog, self).setWindowTitle(title)
        self.lblTitle.setText(title)

    def _setWindowButtonState(self, hint, state):
        btns = {
            Qt.WindowCloseButtonHint: self.btnClose,
            Qt.WindowMinimizeButtonHint: self.btnMinimize,
            Qt.WindowMaximizeButtonHint: self.btnMaximize
        }
        button = btns.get(hint)

        maximized = bool(self.windowState() & Qt.WindowMaximized)

        if button == self.btnMaximize:  # special rules for max/restore
            self.btnRestore.setEnabled(state)
            self.btnMaximize.setEnabled(state)

            if maximized:
                self.btnRestore.setVisible(state)
                self.btnMaximize.setVisible(False)
            else:
                self.btnMaximize.setVisible(state)
                self.btnRestore.setVisible(False)
        else:
            button.setEnabled(state)

        allButtons = [self.btnClose, self.btnMinimize, self.btnMaximize, self.btnRestore]
        if True in [b.isEnabled() for b in allButtons]:
            for b in allButtons:
                b.setVisible(True)
            if maximized:
                self.btnMaximize.setVisible(False)
            else:
                self.btnRestore.setVisible(False)
            self.lblTitle.setContentsMargins(0, 0, 0, 0)
        else:
            for b in allButtons:
                b.setVisible(False)
            self.lblTitle.setContentsMargins(0, 2, 0, 0)

    def setWindowFlag(self, Qt_WindowType, on=True):
        buttonHints = [Qt.WindowCloseButtonHint, Qt.WindowMinimizeButtonHint, Qt.WindowMaximizeButtonHint]

        if Qt_WindowType in buttonHints:
            self._setWindowButtonState(Qt_WindowType, on)
        else:
            QWidget.setWindowFlag(self, Qt_WindowType, on)

    def setWindowFlags(self, Qt_WindowFlags):
        buttonHints = [Qt.WindowCloseButtonHint, Qt.WindowMinimizeButtonHint, Qt.WindowMaximizeButtonHint]
        for hint in buttonHints:
            self._setWindowButtonState(hint, bool(Qt_WindowFlags & hint))

        QWidget.setWindowFlags(self, Qt_WindowFlags)

    @pyqtSlot()
    def on_btnMinimize_clicked(self):
        self.setWindowState(Qt.WindowMinimized)

    @pyqtSlot()
    def on_btnRestore_clicked(self):
        if self.btnMaximize.isEnabled() or self.btnRestore.isEnabled():
            self.btnRestore.setVisible(False)
            self.btnRestore.setEnabled(False)
            self.btnMaximize.setVisible(True)
            self.btnMaximize.setEnabled(True)

        self.setWindowState(Qt.WindowNoState)

    @pyqtSlot()
    def on_btnMaximize_clicked(self):
        if self.btnMaximize.isEnabled() or self.btnRestore.isEnabled():
            self.btnRestore.setVisible(True)
            self.btnRestore.setEnabled(True)
            self.btnMaximize.setVisible(False)
            self.btnMaximize.setEnabled(False)

        self.setWindowState(Qt.WindowMaximized)

    @pyqtSlot()
    def on_btnClose_clicked(self):
        self.close()

    @pyqtSlot()
    def on_titleBar_doubleClicked(self):
        if not bool(self.windowState() & Qt.WindowMaximized):
            self.on_btnMaximize_clicked()
        else:
            self.on_btnRestore_clicked()


class ModernMessageBox(QMessageBox):
    """ Modern message box.
        Args:
            parent (QWidget, optional): Parent widget.
    """
    def __init__(self, parent=None):
        self._dialog = ModernDialog(parent=parent)
        QMessageBox.__init__(self, self._dialog)

        self.setupUi()

    def setupUi(self):
        self.setWindowFlags(Qt.Window | Qt.FramelessWindowHint | Qt.WindowSystemMenuHint)
        self.setAttribute(Qt.WA_TranslucentBackground)

        layout = QHBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self, alignment=Qt.AlignCenter)
        self._dialog.windowContent.setLayout(layout)

        self._dialog.setWindowFlags(Qt.Window | Qt.FramelessWindowHint | Qt.WindowSystemMenuHint)

        self._dialog.show()

    def setWindowTitle(self, title):
        self._dialog.setWindowTitle(title)

    def show(self):
        self._dialog.show()
        super().show()

    def hide(self):
        super().hide()
        self._dialog.hide()



import os

import injector
import loguru

from labbie import constants
from labbie import state
from labbie import utils
from labbie.ui.about.widget import view

logger = loguru.logger
_Constants = constants.Constants


class AboutPresenter:

    @injector.inject
    def __init__(
        self,
        constants: _Constants,
        app_state: state.AppState,
        view: view.AboutWidget
    ):
        self._constants = constants
        self._app_state = app_state
        self._view = view

        self._view.update_relaunch_button(to_debug=not self._constants.debug)
        self.refresh_scrapes()

        self._view.set_relaunch_handler(self.on_relaunch)
        self._view.set_open_data_handler(self.on_open_data)
        self._view.set_open_logs_handler(self.on_open_logs)

        self._app_state.league_enchants.attach(self, self.refresh_scrapes, to='date')
        self._app_state.daily_enchants.attach(self, self.refresh_scrapes, to='date')

    @property
    def widget(self):
        return self._view

    def cleanup(self):
        pass

    def show(self):
        self._view.show()

    def refresh_scrapes(self, *args, **kwargs):
        self._view.set_scrapes(
            league=self._app_state.league_enchants.date,
            daily=self._app_state.daily_enchants.date
        )

    def on_relaunch(self, checked):
        utils.relaunch(debug=not self._constants.debug, exit_fn=lambda: self._view.exit())

    def on_open_data(self, checked):
        if self._constants.data_dir.is_dir():
            os.startfile(self._constants.data_dir)

    def on_open_logs(self, checked):
        if self._constants.logs_dir.is_dir():
            os.startfile(self._constants.logs_dir)



import inspect

import qasync
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets

from labbie.ui import utils


class SystemTrayIcon(QtWidgets.QSystemTrayIcon):
    doubleClicked = QtCore.pyqtSignal()

    def __init__(self):
        super().__init__()
        icon_path = utils.asset_path('taxi.png')
        self.setIcon(QtGui.QIcon(str(icon_path)))
        self.menu = QtWidgets.QMenu()

        self.action_search = self.menu.addAction('Търсене')
        self.action_settings = self.menu.addAction('Настройки')
        self.action_about = self.menu.addAction('За програмата')
        action_exit = self.menu.addAction('Изход')
        action_exit.triggered.connect(self.exit)
        self.setContextMenu(self.menu)

        self.activated.connect(self._on_activated)
        utils.register_exit_handler(self.hide)

    def exit(self):
        QtWidgets.QApplication.exit()

    def _on_activated(self, reason):
        if reason == QtWidgets.QSystemTrayIcon.ActivationReason.DoubleClick:
            self.doubleClicked.emit()

    def set_double_click_handler(self, handler):
        self._connect_signal_to_slot(self.doubleClicked, handler)

    def set_search_triggered_handler(self, handler):
        self._connect_signal_to_slot(self.action_search.triggered, handler)

    def set_settings_triggered_handler(self, handler):
        self._connect_signal_to_slot(self.action_settings.triggered, handler)

    def set_about_triggered_handler(self, handler):
        self._connect_signal_to_slot(self.action_about.triggered, handler)

    @staticmethod
    def _connect_signal_to_slot(signal, slot):
        if inspect.iscoroutinefunction(slot):
            slot = qasync.asyncSlot()(slot)
        signal.connect(slot)



import collections
import copy
import dataclasses
import enum
from typing import Any, Callable, Dict, Iterable, Optional, Sequence, Set, Union

import dacite
import loguru

logger = loguru.logger


class Sentinel(enum.Enum):
    NOT_LOADED = enum.auto()
    NOT_SET = enum.auto()
    NOT_CONFIGURED = enum.auto()
    DEFAULT = enum.auto()


@dataclasses.dataclass
class ObserverHandlers:
    generic_handler: Optional[Callable] = None
    event_handlers: Dict[str, Callable] = dataclasses.field(default_factory=dict)

    def __hash__(self):
        return id(self)


class ObservableMixin:
    def __init__(self):
        self._observer_handlers: Dict[Any, ObserverHandlers] = {}
        self._generic_observer_handlers: Set[ObserverHandlers] = set()
        self._event_observer_handlers: Dict[str, Set[ObserverHandlers]] = collections.defaultdict(set)

    def __post_init__(self):
        if dataclasses.is_dataclass(self):
            try:
                super().__post_init__()
            except AttributeError:
                pass
            self._observer_handlers = {}
            self._generic_observer_handlers = set()
            self._event_observer_handlers = collections.defaultdict(set)

    def attach(self, observer, handler, to: Union[str, Sequence[str]] = None):
        """Attaches a handler and an observer to this observable.

        The provided handler is either attached generically (when `to` is not provided)
        or to a specific event or set of events.
        """
        handlers = self._observer_handlers.setdefault(observer, ObserverHandlers())
        if to is None:
            if handlers.generic_handler is not None:
                logger.warning(f'Overriding generic handler for {observer} observing {self}.')
            handlers.generic_handler = handler
            self._generic_observer_handlers.add(handlers)
        else:
            events = (to, ) if isinstance(to, str) else to
            for event in events:
                if event in handlers.event_handlers:
                    logger.warning(f'Overriding "{event}" event handler for {observer} observing {self}.')
                handlers.event_handlers[event] = handler
                self._event_observer_handlers[event].add(handlers)

    def detatch(self, observer, generic: bool = None, events: Iterable[str] = None):
        observer_handlers = self._observer_handlers.get(observer)
        if observer_handlers:
            if generic is None and events is None:
                # fully detach
                del self._observer_handlers[observer]
                self._generic_observer_handlers.discard(observer_handlers)
                for event in observer_handlers.event_handlers:
                    self._event_observer_handlers[event].discard(observer_handlers)
            else:
                if generic:
                    observer_handlers.generic_handler = None
                    self._generic_observer_handlers.discard(observer_handlers)

                if events:
                    for event in events:
                        self._event_observer_handlers[event].discard(observer_handlers)
                        observer_handlers.event_handlers.pop(event, None)

    def notify(self, _log=True, **kwargs):
        if kwargs:
            for event, args in kwargs.items():
                if not isinstance(args, tuple):
                    args = (args, )
                if _log:
                    logger.debug(f'{self.__class__.__module__}.{self.__class__.__name__} notify, {event=} {args=}')
                for handler in self._event_observer_handlers[event]:
                    handler.event_handlers[event](*args)
        if _log:
            logger.debug(f'{self.__class__.__module__}.{self.__class__.__name__} generic notify')
        # generic notify to call generic handler
        for handler in self._generic_observer_handlers:
            handler.generic_handler(self)


_SERIALIZABLE_CLASSES = {}
_TYPE_HOOKS = {}
_DEFAULT_DACITE_CONFIG = dacite.Config(type_hooks=_TYPE_HOOKS)


def _make_type_hook(cls):
    def type_hook(value):
        logger.debug(f'HOOKED {cls.__module__}.{cls.__name__}')
        if isinstance(value, cls):
            return value
        elif isinstance(value, dict):
            return cls.from_dict(value)
        raise ValueError(f'Can\'t convert {value=} to type {cls.__module__}.{cls.__name__}')
    return type_hook


def _register_serializable_class(cls):
    key = f'{cls.__module__}.{cls.__name__}'
    if key not in _SERIALIZABLE_CLASSES:
        _SERIALIZABLE_CLASSES[key] = cls
        _TYPE_HOOKS[cls] = _make_type_hook(cls)


class SerializableMixin:
    def __init_subclass__(cls, **kwargs):
        _register_serializable_class(cls)
        super().__init_subclass__(**kwargs)

    @staticmethod
    def _dict_factory(result):
        # NOT_LOADED should only be a default value, so we're
        # safe to just remove it
        d = {}
        for k, v in result:
            if k.startswith('_') or v is Sentinel.NOT_LOADED:
                continue
            d[k] = v
        return d

    def as_dict(self):
        d = dataclasses.asdict(self, dict_factory=self._dict_factory)
        if version := self.migration_version():
            d['__VERSION__'] = version
        return d

    @classmethod
    def migration_version(cls):
        v = {}
        classes = cls.__mro__

        # NOTE (bnorick): it's ok to override cls here because we don't need
        # it any longer
        for cls in classes:
            if '_VERSION' in cls.__dict__:
                v[f'{cls.__module__}.{cls.__name__}'] = cls._VERSION
        return v or None

    @classmethod
    def versioned_classes(cls):
        versioned_classes = {}
        for c in cls.__mro__:
            if '_VERSION' in c.__dict__:
                versioned_classes[f'{c.__module__}.{c.__name__}'] = c
        return versioned_classes

    @classmethod
    def migrate(cls, d):
        d_migration_version = d.pop('__VERSION__', None)
        if not d_migration_version:
            return d

        # NOTE (bnorick): I don't think the following logic for combining multiple migrations
        # is not resilient to inheritance changes, if we ever need to do that we'll have to adjust
        # this logic.

        versioned_classes = cls.versioned_classes()
        cur_migration_version = cls.migration_version()

        class_migrated_d = {}
        for class_name, to_v in cur_migration_version.items():
            from_v = d_migration_version[class_name]
            class_migrated = copy.deepcopy(d)
            for v1 in range(from_v, to_v):
                v2 = v1 + 1
                versioned_classes[class_name]._incremental_migration(class_migrated, v1, v2)
            class_migrated_d[class_name] = class_migrated

        # the most specific classes should be combined last, thereby overriding
        # more generic migrations with more specific
        class_order = reversed(versioned_classes.keys())
        migrated_d = class_migrated_d[next(class_order)]
        for class_name in class_order:
            migrated_d.update(class_migrated_d[class_name])

        return migrated_d

    @classmethod
    def _incremental_migration(cls, d, v1, v2):
        migrations = cls.__dict__['_MIGRATIONS']
        migration_fn = getattr(migrations, f'v{v1}_to_v{v2}', None)
        if not migration_fn:
            raise RuntimeError(f'Missing migration from v{v1} to v{v2}.')

        return migration_fn(d)

    @classmethod
    def from_dict(cls, d, **kwargs):
        logger.debug(f'{cls.__module__}.{cls.__name__} __VERSION__={d.get("__VERSION__", "Missing")}')

        d = cls.migrate(d)
        try:
            if config := kwargs.get('config'):
                type_hooks = _TYPE_HOOKS.copy()
                type_hooks.update(config.type_hooks)
                config = copy.deepcopy(config)
                config.type_hooks = type_hooks
                kwargs['config'] = config
            else:
                kwargs['config'] = _DEFAULT_DACITE_CONFIG
            return dacite.from_dict(data_class=cls, data=d, **kwargs)
        except:
            import pprint
            logger.debug(cls)
            logger.debug(pprint.pformat(d))
            raise



import injector
import loguru

from labbie import config
from labbie import constants

logger = loguru.logger


class Module(injector.Module):
    def __init__(self, **kwargs):
        super().__init__()
        self.constants_kwargs = kwargs

    @injector.singleton
    @injector.provider
    def provide_constants(self) -> constants.Constants:
        return constants.Constants.load(**self.constants_kwargs)

    @injector.singleton
    @injector.provider
    def provide_config(self, constants: constants.Constants) -> config.Config:
        loaded = config.Config.load(base_path=constants.config_dir)
        logger.debug(f'Loaded config {loaded}')
        return loaded



import sys
import platform
from os.path import join, dirname, abspath

from PyQt5.QtCore import QT_VERSION_STR

QT_VERSION = tuple(int(v) for v in QT_VERSION_STR.split('.'))
""" tuple: Qt version. """

PLATFORM = platform.system()


def resource_path(relative_path):
    if hasattr(sys, '_MEIPASS'):
        return join(sys._MEIPASS, dirname(abspath(__file__)), relative_path)
    return join(dirname(abspath(__file__)), relative_path)



import injector
import loguru

from labbie import resources

logger = loguru.logger


@injector.singleton
class Trade:

    @injector.inject
    def __init__(self, resource_manager: resources.ResourceManager):
        self._trade_stats = resource_manager.trade_stats

        self.text_to_stat_id = self._trade_stats
